@@ -4,8 +4,10 @@
    Functions that aid in the parsing, serialization, and manipulation
    of PuppetDB queries embedded in HTTP parameters."
   (:require [puppetlabs.puppetdb.cheshire :as json]
-            [clojure.walk :refer [keywordize-keys stringify-keys]]
+            [clojure.java.io]
             [clojure.core.match :as cm]
+            [clojure.tools.logging :as log]
+            [clojure.walk :refer [keywordize-keys stringify-keys]]
             [puppetlabs.puppetdb.query-eng :as qeng]
             [clojure.set :as set]
             [puppetlabs.i18n.core :refer [trs tru]]
@@ -14,13 +16,17 @@
             [puppetlabs.puppetdb.jdbc :as jdbc]
             [puppetlabs.puppetdb.http :as http]
             [puppetlabs.puppetdb.schema :as pls]
-            [puppetlabs.puppetdb.query.paging :refer [parse-limit
+            [ring.util.request :as request]
+            [puppetlabs.puppetdb.query.paging :refer [parse-explain
+                                                      parse-limit
                                                       parse-offset
                                                       parse-order-by
                                                       parse-order-by-json]]
             [puppetlabs.puppetdb.pql :as pql]
             [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [puppetlabs.puppetdb.utils :refer [update-when]]))
+            [puppetlabs.puppetdb.utils :refer [update-when]]
+            [puppetlabs.puppetdb.utils.string-formatter :refer [pprint-json-parse-exception]])
+  (:import (com.fasterxml.jackson.core JsonParseException)))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Schemas
@@ -33,10 +39,14 @@
   {(s/optional-key :query) (s/maybe [s/Any])
    (s/optional-key :ast_only) (s/maybe s/Bool)
    (s/optional-key :include_total) (s/maybe s/Bool)
+   (s/optional-key :optimize_drop_unused_joins) (s/maybe s/Bool)
    (s/optional-key :pretty) (s/maybe s/Bool)
+   (s/optional-key :explain) (s/maybe s/Keyword)
+   (s/optional-key :include_facts_expiration) (s/maybe s/Bool)
    (s/optional-key :include_package_inventory) (s/maybe s/Bool)
    (s/optional-key :order_by) (s/maybe [[(s/one s/Keyword "field")
                                          (s/one (s/enum :ascending :descending) "order")]])
+   (s/optional-key :origin) (s/maybe s/Str)
    (s/optional-key :distinct_resources) (s/maybe s/Bool)
    (s/optional-key :distinct_start_time) s/Any
    (s/optional-key :distinct_end_time) s/Any
@@ -240,21 +250,15 @@
   [params
    param-spec :- param-spec-schema]
   (let [params (stringify-keys params)]
-    (kitchensink/cond-let
-      [p]
-      (kitchensink/excludes-some params (:required param-spec))
+    (when-let [excluded (kitchensink/excludes-some params (:required param-spec))]
       (throw (IllegalArgumentException.
-              (tru "Missing required query parameter ''{0}''" p)))
-
-      (let [diff (set/difference (kitchensink/keyset params)
-                                 (set (:required param-spec))
-                                 (set (:optional param-spec)))]
-        (seq diff))
+              (tru "Missing required query parameter ''{0}''" excluded))))
+    (when-let [invalid (seq (set/difference (kitchensink/keyset params)
+                                            (set (:required param-spec))
+                                            (set (:optional param-spec))))]
       (throw (IllegalArgumentException.
-               (tru "Unsupported query parameter ''{0}''" (first p))))
-
-      :else
-      params)))
+              (tru "Unsupported query parameter ''{0}''" (first invalid)))))
+    params))
 
 (pls/defn-validated convert-query-params :- puppetdb-query-schema
   "This will update a query map to contain the parsed and validated query parameters"
@@ -267,31 +271,89 @@
       (update-when [:limit] parse-limit)
       (update-when [:offset] parse-offset)
       (update-when [:include_total] coerce-to-boolean)
+      (update-when [:optimize_drop_unused_joins] coerce-to-boolean)
       (update-when [:pretty] coerce-to-boolean)
+      (update-when [:explain] parse-explain)
+      (update-when [:include_facts_expiration] coerce-to-boolean)
       (update-when [:include_package_inventory] coerce-to-boolean)
       (update-when [:distinct_resources] coerce-to-boolean)))
 
+(defn parse-json-sequence
+  "Parse a query string as JSON. Parse errors will result in an
+  IllegalArgumentException. Should not be used as a parse-fn directly. Use
+  parse-json-query instead"
+  [query]
+  (try
+    (with-open [string-reader (java.io.StringReader. query)]
+      (doall (json/parsed-seq string-reader)))
+    (catch JsonParseException e
+      (throw (IllegalArgumentException. (pprint-json-parse-exception e query))))))
+
+(defn parse-json-query
+  "Parse a query string as JSON. Multiple queries or any other
+  data, after the query, will result in an IllegalArgumentException"
+  [query query-uuid log-queries?]
+  (when log-queries?
+    (log/info (trs "Parsing PDBQuery:{0}:{1}" query-uuid (pr-str query))))
+  (when query
+    (let [[parsed & others] (parse-json-sequence query)]
+      (when others
+        (throw (IllegalArgumentException.
+                (tru "Only one query may be sent in a request. Found JSON {0} after the query {1}"
+                     others parsed))))
+      parsed)))
+
+(defn time-and-parse-pql
+  "Time how long it takes to transform a PQL query to AST, if the time
+   is greater than 1s, a warning is logged. Returns the transformed
+   query."
+  [query query-uuid log-queries?]
+  (when log-queries?
+    (log/info (trs "Parsing PDBQuery:{0}:{1}" query-uuid (pr-str query))))
+  (let [start (System/nanoTime)
+        result (pql/parse-pql-query query)
+        elapsed (/ (- (System/nanoTime) start) 1000000.0)
+        max-pql-limit 1000]
+    (when (and log-queries? (> elapsed max-pql-limit))
+      (log/warn (trs "Parsing PQL took {0} ms for PDBQuery:{1}:{2}" elapsed query-uuid (pr-str query))))
+    result))
+
+(defn parse-json-or-pql-to-ast
+  "Parse a query string either as JSON or PQL to transform it to AST"
+  [query query-uuid log-queries?]
+  (when query
+    (if (re-find #"^\s*\[" query)
+      (parse-json-query query query-uuid log-queries?)
+      (time-and-parse-pql query query-uuid log-queries?))))
+
 (defn get-req->query
   "Converts parameters of a GET request to a pdb query map"
-  [{:keys [params] :as req}
+  [{:keys [params globals] query-uuid :puppetdb-query-uuid :as _req}
    parse-fn]
-  (-> params
-      (update-when ["query"] parse-fn)
+  (let
+    [log-queries? (:log-queries globals)]
+    (-> params
+      (update-when ["query"] #(parse-fn % query-uuid log-queries?))
       (update-when ["order_by"] parse-order-by-json)
       (update-when ["counts_filter"] json/parse-strict-string true)
       (update-when ["pretty"] coerce-to-boolean)
+      (update-when ["include_facts_expiration"] coerce-to-boolean)
       (update-when ["include_package_inventory"] coerce-to-boolean)
-      keywordize-keys))
+      keywordize-keys)))
 
 (defn post-req->query
   "Takes a POST body and parses the JSON to create a pdb query map"
-  [req parse-fn]
-  (-> (with-open [reader (-> req :body clojure.java.io/reader)]
-        (json/parse-stream reader true))
-      (update :query (fn [query]
-                       (if (vector? query)
-                         query
-                         (parse-fn query))))))
+  [{query-uuid :puppetdb-query-uuid :as req} parse-fn]
+  (let [log-queries? (get-in req [:globals :log-queries])
+        req-body (request/body-string req)
+        parsed-body (try (json/parse-string req-body true)
+                      (catch JsonParseException e
+                        (throw (IllegalArgumentException.
+                                (pprint-json-parse-exception e req-body)))))]
+    (update parsed-body :query (fn [query]
+                                 (if (vector? query)
+                                   query
+                                   (parse-fn query query-uuid log-queries?))))))
 
 (pls/defn-validated create-query-map :- puppetdb-query-schema
   "Takes a ring request map and extracts the puppetdb query from the
@@ -309,22 +371,24 @@
   "Query handler that converts the incoming request (GET or POST)
   parameters/body to a pdb query map"
   ([handler param-spec]
-   (extract-query handler param-spec pql/parse-json-query))
+   (extract-query handler param-spec parse-json-query))
   ([handler param-spec parse-fn]
    (fn [{:keys [puppetdb-query] :as req}]
      (handler
       (if puppetdb-query
         req
-        (let [query-map (create-query-map req param-spec parse-fn)
+        (let [query-uuid (str (java.util.UUID/randomUUID))
+              req-with-query-uuid (assoc req :puppetdb-query-uuid query-uuid)
+              query-map (create-query-map req-with-query-uuid param-spec parse-fn)
               pretty-print (:pretty query-map
                                     (get-in req [:globals :pretty-print]))]
-          (-> req
+          (-> req-with-query-uuid
               (assoc :puppetdb-query query-map)
               (assoc-in [:globals :pretty-print] pretty-print))))))))
 
 (defn extract-query-pql
   [handler param-spec]
-  (extract-query handler param-spec pql/parse-json-or-pql-to-ast))
+  (extract-query handler param-spec parse-json-or-pql-to-ast))
 
 (defn validate-distinct-options!
   "Validate the HTTP query params related to a `distinct_resources` query.  Return a
@@ -364,22 +428,26 @@
 (defn narrow-globals
   "Reduces the number of globals to limit their reach in the codebase"
   [globals]
-  (select-keys globals [:scf-read-db :warn-experimental :url-prefix :pretty-print :node-purge-ttl]))
+  (select-keys globals [:scf-read-db :warn-experimental :url-prefix
+                        :pretty-print :node-purge-ttl :add-agent-report-filter
+                        :log-queries]))
 
 (defn valid-query?
   [scf-read-db version query-map query-options]
-  (let [{:keys [remaining-query entity query-options]}
-        (qeng/user-query->engine-query version query-map (:node-purge-ttl query-options))]
+  (let [options (select-keys query-options [:node-purge-ttl :add-agent-report-filter])
+        {:keys [remaining-query entity query-options]}
+        (qeng/user-query->engine-query version query-map options)]
     (jdbc/with-db-connection scf-read-db
       (when (qeng/query->sql remaining-query entity version query-options)
         true))))
 
 (defn query-handler
   [version]
-  (fn [{:keys [params globals puppetdb-query]}]
+  (fn [{:keys [params globals puppetdb-query puppetdb-query-uuid]}]
     (let [query-options (narrow-globals globals)]
       (if (and (:ast_only puppetdb-query) (valid-query? (:scf-read-db globals) version puppetdb-query query-options))
         (http/json-response (:query puppetdb-query))
         (qeng/produce-streaming-body version
                                 (validate-distinct-options! (merge (keywordize-keys params) puppetdb-query))
+                                puppetdb-query-uuid
                                 query-options)))))
