@@ -103,11 +103,15 @@
 (def ValidX500Name
   (schema/pred valid-x500-name?))
 
-(def SSLExtension
-  "A map containing all the fields required to define an extension."
+(def  SSLExtension
+  "A map containing all the fields required to define an extension. This is not
+   the actual SSL Extension itself but the required oid, critical boolean, and
+   value for the Extension. Use the optional :options to to modify how the
+   extension is created in ExtensionsUtils."
   {:oid schema/Str
    :critical schema/Bool
-   :value Object})
+   :value Object
+   (schema/optional-key :options) Object})
 
 (def SSLExtensionList
   [SSLExtension])
@@ -119,7 +123,7 @@
   (schema/cond-pre PublicKey PrivateKey))
 
 (def ExtensionContainer
-  "A schema for all the things that can hold and SSL extension."
+  "A schema for all the things that can hold an SSL extension."
   (schema/cond-pre
    X509Certificate
    PKCS10CertificationRequest
@@ -128,11 +132,11 @@
 
 (def Readerable
   "Schema for anything that can be handed off to clojure's reader function."
-  (schema/cond-pre Reader BufferedReader InputStream File URI URL Socket byte[] char[] String))
+  (schema/cond-pre Reader BufferedReader InputStream File URI URL Socket bytes chars String))
 
 (def Writerable
   "Schema for anything that can be handed off to clojure's writer function."
-  (schema/cond-pre Writer BufferedWriter OutputStream File URI URL Socket byte[] char[] String))
+  (schema/cond-pre Writer BufferedWriter OutputStream File URI URL Socket bytes chars String))
 
 (def SSLContextOptions
   "Schema for the options map that generate-ssl-context requires."
@@ -202,10 +206,18 @@
   "AuthorityKeyIdentifier OID 2.5.29.35"
   ExtensionsUtils/AUTHORITY_KEY_IDENTIFIER_OID)
 
+(def subject-key-identifier-oid
+  "SubjectKeyIdentifier OID 2.5.29.14"
+  ExtensionsUtils/SUBJECT_KEY_IDENTIFIER_OID)
+
 (def subject-alt-name-oid
   "SubjectAlternativeName OID 2.5.29.17"
   ExtensionsUtils/SUBJECT_ALTERNATIVE_NAME_OID)
 
+(def delta-crl-indicator-oid
+  "DeltaCRLIndicator OID 2.5.29.27"
+  ExtensionsUtils/DELTA_CRL_INDICATOR_OID)
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Extensions
 
@@ -279,27 +291,44 @@
    :value    comment})
 
 (defn- create-authority-key-identifier
-  [public-key issuer-dn serial critical]
+  [{:keys [public-key issuer-dn serial cert critical]}]
   {:oid      authority-key-identifier-oid
    :critical (boolean critical)
    :value    {:public-key    public-key
               :serial-number (if (number? serial) (biginteger serial))
-              :issuer-dn     issuer-dn}})
+              :issuer-dn     issuer-dn
+              :cert          cert}})
 
-(schema/defn ^:always-validate authority-key-identifier :- SSLExtension
-  "Create an `Authority Key Identifier` extension from a `PublicKey` object. The
-  extension created by this function is intended to be passed into
-  `sign-certificate` and `generate-certificate-request`, at which time the key's
-  hash will be computed and stored in the resulting object."
+(schema/defn ^:always-validate authority-key-identifier-options :- SSLExtension
+  "Create an `Authority Key Identifier` extension from a `PublicKey` object or
+   copy the existing `SubjectKeyIdentifier` from a certificate. This function is
+   intended to be passed into `sign-certificate` and `generate-certificate-request`.
+
+   PublicKey details:
+   A type 1 identifier will be computed from the public key.
+
+   Certificate details:
+   An `X509Certificate` can be used instead of a `PublicKey`. The
+   `SubjectKeyIdentifier`of the certificate is copied and used for the
+   `AuthorityKeyIdentifier` for the certificate about to be signed. When a
+   certificate is used, no hash computation takes place."
+  ([cert]
+   (create-authority-key-identifier {:cert cert :critical false}))
   ([public-key critical]
-   (create-authority-key-identifier public-key nil nil critical))
+   (create-authority-key-identifier {:public-key public-key :critical critical}))
   ([issuer-dn serial critical]
-   (create-authority-key-identifier nil issuer-dn serial critical))
+   (create-authority-key-identifier {:issuer-dn issuer-dn :serial serial :critical critical}))
   ([public-key :- PublicKey
     issuer-dn :- ValidX500Name
     serial :- schema/Int
     critical :- Object]
-   (create-authority-key-identifier public-key issuer-dn serial critical)))
+   (create-authority-key-identifier {:public-key public-key
+                                     :issuer-dn issuer-dn
+                                     :serial serial
+                                     :critical critical})))
+
+(schema/def authority-key-identifier
+  authority-key-identifier-options)
 
 (schema/defn ^:always-validate subject-key-identifier :- SSLExtension
   "Create a `Subject Key Identifier` extension from a `PublicKey` object. The
@@ -308,7 +337,7 @@
   hash will be computed and stored in the resulting object."
   [public-key :- PublicKey
    critical :- Object]
-  {:oid      "2.5.29.14"
+  {:oid      subject-key-identifier-oid
    :critical (boolean critical)
    :value    public-key})
 
@@ -353,17 +382,27 @@
 
 (schema/defn create-ca-extensions :- (schema/pred extension-list?)
   "Create a list of extensions to be added to the CA certificate."
-  [ca-name :- (schema/pred valid-x500-name?)
-   ca-serial :- (schema/pred number?)
-   ca-public-key :- (schema/pred public-key?)]
-  [(authority-key-identifier
-     ca-name ca-serial false)
-   (basic-constraints-for-ca)
-   (key-usage
+  ([issuer-public-key :- (schema/pred public-key?)
+    ca-public-key :- (schema/pred public-key?)]
+   [(authority-key-identifier
+     issuer-public-key false)
+    (basic-constraints-for-ca)
+    (key-usage
      #{:key-cert-sign
        :crl-sign} true)
-   (subject-key-identifier
+    (subject-key-identifier
      ca-public-key false)])
+  ([ca-name :- (schema/pred valid-x500-name?)
+    ca-serial :- (schema/pred number?)
+    ca-public-key :- (schema/pred public-key?)]
+   [(authority-key-identifier
+     ca-name ca-serial false)
+    (basic-constraints-for-ca)
+    (key-usage
+     #{:key-cert-sign
+       :crl-sign} true)
+    (subject-key-identifier
+     ca-public-key false)]))
 
 (schema/defn ^:always-validate crl-number :- SSLExtension
   "Create a `CRL Number` extension"
@@ -372,6 +411,12 @@
    :critical false
    :value (biginteger number)})
 
+(schema/defn ^:always-validate get-crl-number :- (schema/maybe BigInteger)
+  "Given a CRL, return the value of the CRL Number extension, or nil if the
+  extension is not present."
+  [crl :- X509CRL]
+  (get-extension-value crl crl-number-oid))
+
 (schema/defn ^:always-validate puppet-node-uid :- SSLExtension
   "Create a `Puppet Node UID` extension."
   [uid :- schema/Str
@@ -404,6 +449,17 @@
    :critical (boolean critical)
    :value key})
 
+(schema/defn ^:always-validate delta-crl-indicator :- SSLExtension
+  "Create a `Delta CRL Indicator` extension."
+  [base-crl-number :- BigInteger]
+  {:oid delta-crl-indicator-oid
+   :critical true
+   :value base-crl-number})
+
+(schema/defn ^:always-validate delta-crl? :- schema/Bool
+  [crl :- X509CRL]
+  (boolean (get-extension-value crl delta-crl-indicator-oid)))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Core
 
@@ -492,16 +548,31 @@
 (schema/defn ^:always-validate generate-crl :- X509CRL
   "Given the certificate authority's principal identifier and private & public
    keys, create and return a `X509CRL` certificate revocation list (CRL).
-   The CRL will have an AuthorityKeyIdentifier and CRLNumber extensions.
+   The CRL will have an AuthorityKeyIdentifier and CRLNumber extensions. Typical
+   usage of this will be with the 3 argument arity; the fuller arity is exposed
+   here for testing purposes.
 
    Arguments:
    `issuer`:             the issuer's `X500Principal`
    `issuer-private-key`: the issuer's `PrivateKey`
-   `issuer-public-key`:  the issuer's `PublicKey`"
-  [issuer :- X500Principal
-   issuer-private-key :- PrivateKey
-   issuer-public-key :- PublicKey]
-  (SSLUtils/generateCRL issuer issuer-private-key issuer-public-key))
+   `issuer-public-key`:  the issuer's `PublicKey`
+   `this-update`:        date for when this crl is created
+   `next-update`:        when to fetch the rcl next
+   `crl-number`:         number to start revocation count at
+   `extensions`:         any extensions to sign onto the crl"
+  ([issuer :- X500Principal
+    issuer-private-key :- PrivateKey
+    issuer-public-key :- PublicKey]
+   (SSLUtils/generateCRL issuer issuer-private-key issuer-public-key))
+  ([issuer :- X500Principal
+    issuer-private-key :- PrivateKey
+    issuer-public-key :- PublicKey
+    this-update :- Date
+    next-update :- Date
+    crl-number :- BigInteger
+    extensions :- SSLExtensionList]
+   (SSLUtils/generateCRL issuer issuer-private-key issuer-public-key this-update next-update
+                         crl-number (javaize extensions))))
 
 (schema/defn ^:always-validate revoked? :- schema/Bool
   "Given a certificate revocation list and certificate, test if the
@@ -551,6 +622,15 @@
   (SSLUtils/revokeMultiple crl issuer-private-key
                    issuer-public-key cert-serials))
 
+(schema/defn ^:always-validate validate-cert-chain
+  "Given a list of certificates and a list of CRLs, validate the certificate
+   chain, i.e. ensure that none of the certs have been revoked by checking the
+   appropriate CRL, which must be present and currently valid.
+   Returns nil if successful."
+  [cert-chain :- [X509Certificate]
+   crl-chain :- [X509CRL]]
+  (SSLUtils/validateCertChain cert-chain crl-chain))
+
 (schema/defn ^:always-validate crl->pem!
   "Encodes a CRL to PEM format, and writes it to a file (or other stream).
    Arguments:
@@ -579,13 +659,25 @@
   (with-open [r (reader pem)]
     (SSLUtils/pemToCRLs r)))
 
+(schema/defn ^:always-validate crl-issued-by-cert? :- schema/Bool
+  "Given a CRL and a certificate, determine whether the CRL was issued by the
+  certificate."
+  [crl :- X509CRL
+   cert :- X509Certificate]
+  (= (.getIssuerX500Principal crl)
+     (.getSubjectX500Principal cert)))
+
 (schema/defn ^:always-validate pem->ca-crl :- X509CRL
-  "Given a CRL chain, extract the first CRL"
-  [crl-chain :- Readerable]
-  (let [[crl rest-of-chain] (pem->crls crl-chain)]
+  "Given a CRL chain and CA certificate, extract the CRL issued by the
+  certificate"
+  [crl-chain :- Readerable
+   ca-cert :- X509Certificate]
+  (let [match-ca-cert? (fn [crl] (crl-issued-by-cert? crl ca-cert))
+        crls (pem->crls crl-chain)
+        crl (first (filter match-ca-cert? crls))]
     (if (nil? crl)
       (throw (IllegalArgumentException.
-               "The CRL reader must contain at least one CRL"))
+               "The CRL reader does not contain a CRL matching the given certificate"))
       crl)))
 
 (schema/defn ^:always-validate pem->csr :- PKCS10CertificationRequest
@@ -645,13 +737,13 @@
     (SSLUtils/pemToCerts r)))
 
 (schema/defn ^:always-validate pem->ca-cert :- X509Certificate
-  "Given a CA certificate chain and key pair, extract the first certificate and
-  verify that it matches the key pair."
+  "Given a CA certificate chain and key pair, extract and verify the certificate
+  matching the key pair."
   [cert-chain-pem :- Readerable
-   keypair-pem :- Readerable]
+   key-or-keypair-pem :- Readerable]
   (with-open [cert-chain-pem-reader (reader cert-chain-pem)
-              keypair-pem-reader (reader keypair-pem)]
-    (SSLUtils/pemToCaCert cert-chain-pem-reader keypair-pem-reader)))
+              key-pem-reader (reader key-or-keypair-pem)]
+    (SSLUtils/pemToCaCert cert-chain-pem-reader key-pem-reader)))
 
 (schema/defn ^:always-validate pem->cert :- X509Certificate
   "Given the path to a PEM file (or some other object supported by clojure's `reader`),
