@@ -48,6 +48,49 @@
   (:import (clojure.lang IObj RT Compiler Var)
            java.net.URL))
 
+(def ns-safe-macro
+  "Clojure macros that are known to not alter namespaces"
+  #{#'clojure.core/->
+    #'clojure.core/->>
+    #'clojure.core/..
+    #'clojure.core/and
+    #'clojure.core/as->
+    #'clojure.core/assert
+    #'clojure.core/case
+    #'clojure.core/cond
+    #'clojure.core/cond->
+    #'clojure.core/cond->>
+    #'clojure.core/condp
+    #'clojure.core/defn
+    #'clojure.core/defn-
+    #'clojure.core/delay
+    #'clojure.core/doseq
+    #'clojure.core/dosync
+    #'clojure.core/dotimes
+    #'clojure.core/doto
+    #'clojure.core/fn
+    #'clojure.core/for
+    #'clojure.core/future
+    #'clojure.core/if-let
+    #'clojure.core/if-not
+    #'clojure.core/lazy-seq
+    #'clojure.core/let
+    #'clojure.core/letfn
+    #'clojure.core/loop
+    #'clojure.core/or
+    #'clojure.core/reify
+    #'clojure.core/some->
+    #'clojure.core/some->>
+    #'clojure.core/sync
+    #'clojure.core/time
+    #'clojure.core/when
+    #'clojure.core/when-let
+    #'clojure.core/when-not
+    #'clojure.core/while
+    #'clojure.core/with-open
+    #'clojure.core/with-out-str
+    })
+
 (def specials
   "Set of the special forms for clojure in the JVM"
   (into ana/specials
@@ -67,6 +110,7 @@
 
 (defn global-env []
   (atom {:namespaces     (build-ns-map)
+
          :update-ns-map! (fn update-ns-map! []
                            (swap! *env* assoc-in [:namespaces] (build-ns-map)))}))
 
@@ -146,7 +190,8 @@
 
                macro?
                (let [res (apply v form (:locals env) (rest form))] ; (m &form &env & args)
-                 (update-ns-map!)
+                 (when-not (ns-safe-macro v)
+                    (update-ns-map!))
                  (if (obj? res)
                    (vary-meta res merge (meta form))
                    res))
@@ -240,7 +285,7 @@
                         (not (vector? args))
                         (str "Parameter listing should be a vector, had: " (class args))
                         (not (first args))
-                        (str"Must supply at least one argument for 'this' in: " method))]
+                        (str "Must supply at least one argument for 'this' in: " method))]
     (throw (ex-info error-msg
                     (merge {:form     form
                             :in       (:this env)
@@ -268,14 +313,13 @@
 (defn -deftype [name class-name args interfaces]
 
   (doseq [arg [class-name name]]
-    (memo-clear! maybe-class-from-string [(str arg)])
     (memo-clear! members* [arg])
     (memo-clear! members* [(str arg)]))
 
   (let [interfaces (mapv #(symbol (.getName ^Class %)) interfaces)]
-    (eval (list 'let []
+    (eval (list `let []
                 (list 'deftype* name class-name args :implements interfaces)
-                (list 'import class-name)))))
+                (list `import class-name)))))
 
 (defn parse-reify*
   [[_ interfaces & methods :as form] env]
@@ -464,23 +508,36 @@
                             #'*ns*              (the-ns (:ns env))}
                            (:bindings opts))
        (env/ensure (global-env)
-         (env/with-env (swap! env/*env* mmerge
-                              {:passes-opts (get opts :passes-opts default-passes-opts)})
-           (run-passes (-analyze form env)))))))
+         (doto (env/with-env (mmerge (env/deref-env)
+                                     {:passes-opts (get opts :passes-opts default-passes-opts)})
+                 (run-passes (-analyze form env)))
+           (do (update-ns-map!)))))))
 
-(deftype ExceptionThrown [e])
+(deftype ExceptionThrown [e ast])
+
+(defn ^:private throw! [e]
+  (throw (.e ^ExceptionThrown e)))
 
 (defn analyze+eval
   "Like analyze but evals the form after the analysis and attaches the
    returned value in the :result field of the AST node.
+
    If evaluating the form will cause an exception to be thrown, the exception
-   will be caught and the :result field will hold an ExceptionThrown instance
-   with the exception in the \"e\" field.
+   will be caught and wrapped in an ExceptionThrown object, containing the
+   exception in the `e` field and the AST in the `ast` field.
+
+   The ExceptionThrown object is then passed to `handle-evaluation-exception`,
+   which by defaults throws the original exception, but can be used to provide
+   a replacement return value for the evaluation of the AST.
+
+   Unrolls `do` forms to handle the Gilardi scenario.
 
    Useful when analyzing whole files/namespaces."
   ([form] (analyze+eval form (empty-env) {}))
   ([form env] (analyze+eval form env {}))
-  ([form env opts]
+  ([form env {:keys [handle-evaluation-exception]
+              :or {handle-evaluation-exception throw!}
+              :as opts}]
      (env/ensure (global-env)
        (update-ns-map!)
        (let [env (merge env (-source-info form env))
@@ -492,7 +549,7 @@
                                      (if (= mform form)
                                        [mform (seq raw-forms)]
                                        (recur mform (conj raw-forms
-                                                          (if-let [[op & r] (and (seq? form)form )]
+                                                          (if-let [[op & r] (and (seq? form) form)]
                                                             (if (or (u/macro? op  env)
                                                                     (u/inline? op r env))
                                                               (vary-meta form assoc ::ana/resolved-op (resolve-sym op env))
@@ -520,7 +577,7 @@
                  frm (emit-form a)
                  result (try (eval frm) ;; eval the emitted form rather than directly the form to avoid double macroexpansion
                              (catch Exception e
-                               (ExceptionThrown. e)))]
+                               (handle-evaluation-exception (ExceptionThrown. e a))))]
              (merge a {:result    result
                        :raw-forms raw-forms})))))))
 
